<?php

use PayPalCheckoutSdk\Core\PayPalHttpClient;
use PayPalCheckoutSdk\Core\ProductionEnvironment;
use PayPalCheckoutSdk\Core\SandboxEnvironment;
use PayPalCheckoutSdk\Orders\OrdersCaptureRequest;
use PayPalCheckoutSdk\Orders\OrdersCreateRequest;
use PayPalCheckoutSdk\Orders\OrdersGetRequest;

App::uses('PaypalLib', 'EvPaypal.Lib');
App::uses('IdentityGenerateTokenRequest', 'EvPaypal.Requests');
App::uses('GatewayBaseComponent', 'EvTransactions.Controller/Component');
App::uses('GatewayInterface', 'EvTransactions.Lib');

class PaypalOnsiteComponent extends GatewayBaseComponent implements GatewayInterface {

/**
 * @var array The config for the current environment
 */
	protected $_config;

/**
 * @var PayPalHttpClient The API instance
 */
	protected $_api;

/**
 * {@inheritDoc}
 */
	public function setup() {
		$this->_config = PaypalLib::getConfig();

		if ($this->_config['testMode']) {
			$this->_api = new PayPalHttpClient(
				new SandboxEnvironment($this->_config['clientId'], $this->_config['clientSecret'])
			);
		} else {
			$this->_api = new PayPalHttpClient(
				new ProductionEnvironment($this->_config['clientId'], $this->_config['clientSecret'])
			);
		}
	}

/**
 * Sets the client token to the view data
 *
 * @return void
 */
	public function setClientToken() {
		// Not called in the usual gateway flow so need to call setup manually
		$this->setup();

		// Generate a new hash salt for this request if it is not posting a previous request.
		$hashSalt = PaypalLib::getHashSalt(empty($this->_controller->request->data['Paypal']));
		$request = new IdentityGenerateTokenRequest();
		$response = $this->_api->execute($request);

		if (!empty($response->result->client_token)) {
			$this->_controller->set('paypal', [
				'clientToken' => $response->result->client_token,
				'clientId' => $this->_config['clientId'],
				'brandName' => $this->_config['brandName'],
				'hashSalt' => $hashSalt,
			]);
			return;
		}

		throw new Exception('Payment token could not be generated');
	}

/**
 * Creates a paypal order using the REST API and returns the ID
 *
 * @return void
 */
	public function createOrder($orderData) {
		$request = new OrdersCreateRequest();
		$request->body = $orderData;
		$response = $this->_api->execute($request);

		if ($response->result->status === 'CREATED') {
			return $response->result->id;
		}

		throw new InternalErrorException('Order could not be created');
	}

/**
 * {@inheritDoc}
 */
	public function setupPayment($transactionId, $return, $model, $amount, $items, $extra = array()) {
		$this->_data = [
			'return' => $return,
			'orderId' => $extra['auth']['orderId'],
			'amount' => $amount,
			'hash' => $extra['auth']['hash']
		];
	}

/**
 * {@inheritDoc}
 */
	public function getPayment($transactionId) {
		$this->Transaction = EvClassRegistry::init('EvTransactions.Transaction');

		try {
			// Check the order hash and the 3ds state of the order
			if (!$this->_checkPaymentHash() || !$this->_check3DS()) {
				return [
					'status' => 'failed',
					'result' => false,
				];
			};

			$request = new OrdersCaptureRequest($this->_data['orderId']);
			$response = $this->_api->execute($request);

			if ($response->result->status === 'COMPLETED') {
				$this->Transaction->save([
					'Transaction' => [
						'payment_token' => $response->result->id,
						'status' => 'success',
					],
				]);
				return [
					'status' => 'success',
					'result' => true,
				];
			}
		} catch (Exception $e) {
			$this->Transaction->save([
				'Transaction' => [
					'status' => 'failed',
					'message' => $e->getMessage(),
				],
			]);
			return [
				'status' => 'failed',
				'result' => false,
			];
		}

		$this->Transaction->save([
			'Transaction' => [
				'payment_token' => $response->result->id,
				'status' => 'failed',
				'message' => 'Unrecognised status in response'
			],
		]);
		return [
			'status' => 'failed',
			'result' => false,
		];
	}

/**
 * {@inheritDoc}
 */
	public function processReturn() {
		// Check the status of the transaction for the result
		$this->Transaction = EvClassRegistry::init('EvTransactions.Transaction');
		$transaction = $this->Transaction->findById($this->_controller->request->query['transaction']);

		return [
			'result' => $transaction['Transaction']['status'] === 'success'
		];
	}

/**
 * Undocumented function
 *
 * @return void
 */
	protected function _check3DS() {
		$request = new OrdersGetRequest($this->_data['orderId']);
		$response = $this->_api->execute($request);

		if (!empty($response->result->payment_source->card)) {
			// Paying by card. Check the 3ds status
			$authentication = $response->result->payment_source->card->authentication_result;

			if (!(
				// If NOT authenticatication has succeeded
				$authentication->liability_shift === 'POSSIBLE' ||
				// or card is not enrolled
				($authentication->liability_shift === 'NO' && $authentication->three_d_secure->enrollment_status !== 'Y')
			)) {

				$this->Transaction->save([
					'Transaction' => [
						'payment_token' => $response->result->id,
						'status' => 'failed',
						'message' => '3DS validation failed'
					],
				]);
				return false;
			}
		}

		return true;
	}

	protected function _checkPaymentHash() {
		// Check the hash matches and nobody is fiddling the amounts
		if (is_array($this->_data['amount'])) {
			$total = $this->_data['amount']['amount'];
			$currency = $this->_data['amount']['currency'];
		} else {
			$config = PaypalLib::getConfig();
			$total = $this->_data['amount'];
			$currency = $config['currency'];
		}

		// If the hash doesn't match record a failed transaction
		if (!PaypalLib::checkHash(number_format($total, 2), $currency, $this->_data['hash'])) {
			$this->Transaction = EvClassRegistry::init('EvTransactions.Transaction');
			$this->Transaction->save([
				'Transaction' => [
					'payment_token' => $this->_data['orderId'],
					'status' => 'failed',
					'message' => 'Payment hash check failed'
				],
			]);

			return false;
		}

		return true;
	}
}
